// Copyright 2025 The Sigstore Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.6
// 	protoc        v5.29.3
// source: rekor/v2/dsse.proto

package protobuf

import (
	v1 "github.com/sigstore/protobuf-specs/gen/pb-go/common/v1"
	dsse "github.com/sigstore/protobuf-specs/gen/pb-go/dsse"
	_ "google.golang.org/genproto/googleapis/api/annotations"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

// A request to add a DSSE v0.0.2 entry to the log
type DSSERequestV002 struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// A DSSE envelope
	Envelope *dsse.Envelope `protobuf:"bytes,1,opt,name=envelope,proto3" json:"envelope,omitempty"`
	// All necessary verification material to verify all signatures embedded in the envelope
	Verifiers     []*Verifier `protobuf:"bytes,2,rep,name=verifiers,proto3" json:"verifiers,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DSSERequestV002) Reset() {
	*x = DSSERequestV002{}
	mi := &file_rekor_v2_dsse_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DSSERequestV002) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DSSERequestV002) ProtoMessage() {}

func (x *DSSERequestV002) ProtoReflect() protoreflect.Message {
	mi := &file_rekor_v2_dsse_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DSSERequestV002.ProtoReflect.Descriptor instead.
func (*DSSERequestV002) Descriptor() ([]byte, []int) {
	return file_rekor_v2_dsse_proto_rawDescGZIP(), []int{0}
}

func (x *DSSERequestV002) GetEnvelope() *dsse.Envelope {
	if x != nil {
		return x.Envelope
	}
	return nil
}

func (x *DSSERequestV002) GetVerifiers() []*Verifier {
	if x != nil {
		return x.Verifiers
	}
	return nil
}

type DSSELogEntryV002 struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// The hash of the DSSE payload
	PayloadHash *v1.HashOutput `protobuf:"bytes,1,opt,name=payloadHash,proto3" json:"payloadHash,omitempty"`
	// Signatures and their associated verification material used to verify the payload
	Signatures    []*Signature `protobuf:"bytes,2,rep,name=signatures,proto3" json:"signatures,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *DSSELogEntryV002) Reset() {
	*x = DSSELogEntryV002{}
	mi := &file_rekor_v2_dsse_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *DSSELogEntryV002) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*DSSELogEntryV002) ProtoMessage() {}

func (x *DSSELogEntryV002) ProtoReflect() protoreflect.Message {
	mi := &file_rekor_v2_dsse_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use DSSELogEntryV002.ProtoReflect.Descriptor instead.
func (*DSSELogEntryV002) Descriptor() ([]byte, []int) {
	return file_rekor_v2_dsse_proto_rawDescGZIP(), []int{1}
}

func (x *DSSELogEntryV002) GetPayloadHash() *v1.HashOutput {
	if x != nil {
		return x.PayloadHash
	}
	return nil
}

func (x *DSSELogEntryV002) GetSignatures() []*Signature {
	if x != nil {
		return x.Signatures
	}
	return nil
}

var File_rekor_v2_dsse_proto protoreflect.FileDescriptor

const file_rekor_v2_dsse_proto_rawDesc = "" +
	"\n" +
	"\x13rekor/v2/dsse.proto\x12\x15dev.sigstore.rekor.v2\x1a\x1fgoogle/api/field_behavior.proto\x1a\x15sigstore_common.proto\x1a\x0eenvelope.proto\x1a\x17rekor/v2/verifier.proto\"\x8b\x01\n" +
	"\x0fDSSERequestV002\x124\n" +
	"\benvelope\x18\x01 \x01(\v2\x13.io.intoto.EnvelopeB\x03\xe0A\x02R\benvelope\x12B\n" +
	"\tverifiers\x18\x02 \x03(\v2\x1f.dev.sigstore.rekor.v2.VerifierB\x03\xe0A\x02R\tverifiers\"\xa4\x01\n" +
	"\x10DSSELogEntryV002\x12I\n" +
	"\vpayloadHash\x18\x01 \x01(\v2\".dev.sigstore.common.v1.HashOutputB\x03\xe0A\x02R\vpayloadHash\x12E\n" +
	"\n" +
	"signatures\x18\x02 \x03(\v2 .dev.sigstore.rekor.v2.SignatureB\x03\xe0A\x02R\n" +
	"signaturesBz\n" +
	"\x1bdev.sigstore.proto.rekor.v2B\vRekorV2DsseP\x01Z6github.com/sigstore/rekor-tiles/pkg/generated/protobuf\xea\x02\x13Sigstore::Rekor::V2b\x06proto3"

var (
	file_rekor_v2_dsse_proto_rawDescOnce sync.Once
	file_rekor_v2_dsse_proto_rawDescData []byte
)

func file_rekor_v2_dsse_proto_rawDescGZIP() []byte {
	file_rekor_v2_dsse_proto_rawDescOnce.Do(func() {
		file_rekor_v2_dsse_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_rekor_v2_dsse_proto_rawDesc), len(file_rekor_v2_dsse_proto_rawDesc)))
	})
	return file_rekor_v2_dsse_proto_rawDescData
}

var file_rekor_v2_dsse_proto_msgTypes = make([]protoimpl.MessageInfo, 2)
var file_rekor_v2_dsse_proto_goTypes = []any{
	(*DSSERequestV002)(nil),  // 0: dev.sigstore.rekor.v2.DSSERequestV002
	(*DSSELogEntryV002)(nil), // 1: dev.sigstore.rekor.v2.DSSELogEntryV002
	(*dsse.Envelope)(nil),    // 2: io.intoto.Envelope
	(*Verifier)(nil),         // 3: dev.sigstore.rekor.v2.Verifier
	(*v1.HashOutput)(nil),    // 4: dev.sigstore.common.v1.HashOutput
	(*Signature)(nil),        // 5: dev.sigstore.rekor.v2.Signature
}
var file_rekor_v2_dsse_proto_depIdxs = []int32{
	2, // 0: dev.sigstore.rekor.v2.DSSERequestV002.envelope:type_name -> io.intoto.Envelope
	3, // 1: dev.sigstore.rekor.v2.DSSERequestV002.verifiers:type_name -> dev.sigstore.rekor.v2.Verifier
	4, // 2: dev.sigstore.rekor.v2.DSSELogEntryV002.payloadHash:type_name -> dev.sigstore.common.v1.HashOutput
	5, // 3: dev.sigstore.rekor.v2.DSSELogEntryV002.signatures:type_name -> dev.sigstore.rekor.v2.Signature
	4, // [4:4] is the sub-list for method output_type
	4, // [4:4] is the sub-list for method input_type
	4, // [4:4] is the sub-list for extension type_name
	4, // [4:4] is the sub-list for extension extendee
	0, // [0:4] is the sub-list for field type_name
}

func init() { file_rekor_v2_dsse_proto_init() }
func file_rekor_v2_dsse_proto_init() {
	if File_rekor_v2_dsse_proto != nil {
		return
	}
	file_rekor_v2_verifier_proto_init()
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_rekor_v2_dsse_proto_rawDesc), len(file_rekor_v2_dsse_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   2,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_rekor_v2_dsse_proto_goTypes,
		DependencyIndexes: file_rekor_v2_dsse_proto_depIdxs,
		MessageInfos:      file_rekor_v2_dsse_proto_msgTypes,
	}.Build()
	File_rekor_v2_dsse_proto = out.File
	file_rekor_v2_dsse_proto_goTypes = nil
	file_rekor_v2_dsse_proto_depIdxs = nil
}
