// Code generated by smithy-go-codegen DO NOT EDIT.

// Package ecr provides the API client, operations, and parameter types for Amazon
// Elastic Container Registry.
//
// # Amazon Elastic Container Registry
//
// Amazon Elastic Container Registry (Amazon ECR) is a managed container image
// registry service. Customers can use the familiar Docker CLI, or their preferred
// client, to push, pull, and manage images. Amazon ECR provides a secure,
// scalable, and reliable registry for your Docker or Open Container Initiative
// (OCI) images. Amazon ECR supports private repositories with resource-based
// permissions using IAM so that specific users or Amazon EC2 instances can access
// repositories and images.
//
// Amazon ECR has service endpoints in each supported Region. For more
// information, see [Amazon ECR endpoints]in the Amazon Web Services General Reference.
//
// [Amazon ECR endpoints]: https://docs.aws.amazon.com/general/latest/gr/ecr.html
package ecr
