// Code generated by "stringer -type=tokType"; DO NOT EDIT.

package jmespath

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[tUnknown-0]
	_ = x[tStar-1]
	_ = x[tDot-2]
	_ = x[tFilter-3]
	_ = x[tFlatten-4]
	_ = x[tLparen-5]
	_ = x[tRparen-6]
	_ = x[tLbracket-7]
	_ = x[tRbracket-8]
	_ = x[tLbrace-9]
	_ = x[tRbrace-10]
	_ = x[tOr-11]
	_ = x[tPipe-12]
	_ = x[tNumber-13]
	_ = x[tUnquotedIdentifier-14]
	_ = x[tQuotedIdentifier-15]
	_ = x[tComma-16]
	_ = x[tColon-17]
	_ = x[tLT-18]
	_ = x[tLTE-19]
	_ = x[tGT-20]
	_ = x[tGTE-21]
	_ = x[tEQ-22]
	_ = x[tNE-23]
	_ = x[tJSONLiteral-24]
	_ = x[tStringLiteral-25]
	_ = x[tCurrent-26]
	_ = x[tExpref-27]
	_ = x[tAnd-28]
	_ = x[tNot-29]
	_ = x[tEOF-30]
}

const _tokType_name = "tUnknowntStartDottFiltertFlattentLparentRparentLbrackettRbrackettLbracetRbracetOrtPipetNumbertUnquotedIdentifiertQuotedIdentifiertCommatColontLTtLTEtGTtGTEtEQtNEtJSONLiteraltStringLiteraltCurrenttExpreftAndtNottEOF"

var _tokType_index = [...]uint8{0, 8, 13, 17, 24, 32, 39, 46, 55, 64, 71, 78, 81, 86, 93, 112, 129, 135, 141, 144, 148, 151, 155, 158, 161, 173, 187, 195, 202, 206, 210, 214}

func (i tokType) String() string {
	if i < 0 || i >= tokType(len(_tokType_index)-1) {
		return "tokType(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _tokType_name[_tokType_index[i]:_tokType_index[i+1]]
}
