// Code generated by cuelang.org/go/pkg/gen. DO NOT EDIT.

package template

import (
	"cuelang.org/go/internal/core/adt"
	"cuelang.org/go/internal/pkg"
)

func init() {
	pkg.Register("text/template", p)
}

var _ = adt.TopKind // in case the adt package isn't used

var p = &pkg.Package{
	Native: []*pkg.Builtin{{
		Name: "Execute",
		Params: []pkg.Param{
			{Kind: adt.StringKind},
			{Kind: adt.TopKind},
		},
		Result: adt.StringKind,
		Func: func(c *pkg.CallCtxt) {
			templ, data := c.String(0), c.Value(1)
			if c.Do() {
				c.Ret, c.Err = Execute(templ, data)
			}
		},
	}, {
		Name: "HTMLEscape",
		Params: []pkg.Param{
			{Kind: adt.StringKind},
		},
		Result: adt.StringKind,
		Func: func(c *pkg.CallCtxt) {
			s := c.String(0)
			if c.Do() {
				c.Ret = HTMLEscape(s)
			}
		},
	}, {
		Name: "JSEscape",
		Params: []pkg.Param{
			{Kind: adt.StringKind},
		},
		Result: adt.StringKind,
		Func: func(c *pkg.CallCtxt) {
			s := c.String(0)
			if c.Do() {
				c.Ret = JSEscape(s)
			}
		},
	}},
}
