// Copyright 2019 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package registry

import (
	tinkpb "github.com/tink-crypto/tink-go/v2/proto/tink_go_proto"
)

// PrivateKeyManager is a special type of KeyManager that understands private key types.
type PrivateKeyManager interface {
	KeyManager

	// PublicKeyData extracts the public key data from the private key.
	PublicKeyData(serializedKey []byte) (*tinkpb.KeyData, error)
}
