// Code generated by "stringer -type=ErrorCode -linecomment"; DO NOT EDIT.

package adt

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[EvalError-0]
	_ = x[UserError-1]
	_ = x[StructuralCycleError-2]
	_ = x[IncompleteError-3]
	_ = x[CycleError-4]
}

const _ErrorCode_name = "evaluserstructural cycleincompletecycle"

var _ErrorCode_index = [...]uint8{0, 4, 8, 24, 34, 39}

func (i ErrorCode) String() string {
	if i < 0 || i >= ErrorCode(len(_ErrorCode_index)-1) {
		return "ErrorCode(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _ErrorCode_name[_ErrorCode_index[i]:_ErrorCode_index[i+1]]
}
