// Code generated by "stringer -type=depKind"; DO NOT EDIT.

package adt

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[PARENT-1]
	_ = x[ARC-2]
	_ = x[NOTIFY-3]
	_ = x[TASK-4]
	_ = x[DISJUNCT-5]
	_ = x[EVAL-6]
	_ = x[COMP-7]
	_ = x[ROOT-8]
	_ = x[INIT-9]
	_ = x[DEFER-10]
	_ = x[SHARED-11]
	_ = x[TEST-12]
}

const _depKind_name = "PARENTARCNOTIFYTASKDISJUNCTEVALCOMPROOTINITDEFERSHAREDTEST"

var _depKind_index = [...]uint8{0, 6, 9, 15, 19, 27, 31, 35, 39, 43, 48, 54, 58}

func (i depKind) String() string {
	i -= 1
	if i < 0 || i >= depKind(len(_depKind_index)-1) {
		return "depKind(" + strconv.FormatInt(int64(i+1), 10) + ")"
	}
	return _depKind_name[_depKind_index[i]:_depKind_index[i+1]]
}
