// Code generated by cuelang.org/go/pkg/gen. DO NOT EDIT.

package sha256

import (
	"cuelang.org/go/internal/core/adt"
	"cuelang.org/go/internal/pkg"
)

func init() {
	pkg.Register("crypto/sha256", p)
}

var _ = adt.TopKind // in case the adt package isn't used

var p = &pkg.Package{
	Native: []*pkg.Builtin{{
		Name:  "Size",
		Const: "32",
	}, {
		Name:  "Size224",
		Const: "28",
	}, {
		Name:  "BlockSize",
		Const: "64",
	}, {
		Name: "Sum256",
		Params: []pkg.Param{
			{Kind: adt.BytesKind | adt.StringKind},
		},
		Result: adt.BytesKind | adt.StringKind,
		Func: func(c *pkg.CallCtxt) {
			data := c.Bytes(0)
			if c.Do() {
				c.Ret = Sum256(data)
			}
		},
	}, {
		Name: "Sum224",
		Params: []pkg.Param{
			{Kind: adt.BytesKind | adt.StringKind},
		},
		Result: adt.BytesKind | adt.StringKind,
		Func: func(c *pkg.CallCtxt) {
			data := c.Bytes(0)
			if c.Do() {
				c.Ret = Sum224(data)
			}
		},
	}},
}
