// Copyright 2021 CUE Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package types

import (
	"cuelang.org/go/internal/core/adt"
	"cuelang.org/go/internal/core/runtime"
)

type Value struct {
	R *runtime.Runtime
	V *adt.Vertex
}

type Interface interface {
	Core(v *Value)
}

func CastValue(t *Value, x interface{}) bool {
	c, ok := x.(Interface)
	if ok {
		c.Core(t)
	}
	return ok
}
