// Code generated by "stringer -type=RelPos -linecomment"; DO NOT EDIT.

package token

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[NoRelPos-0]
	_ = x[Elided-1]
	_ = x[NoSpace-2]
	_ = x[Blank-3]
	_ = x[Newline-4]
	_ = x[NewSection-5]
}

const _RelPos_name = "invalidelidednospaceblanknewlinesection"

var _RelPos_index = [...]uint8{0, 7, 13, 20, 25, 32, 39}

func (i RelPos) String() string {
	if i < 0 || i >= RelPos(len(_RelPos_index)-1) {
		return "RelPos(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _RelPos_name[_RelPos_index[i]:_RelPos_index[i+1]]
}
