// Package ordered implements an ordered map type, suitable for use when decoding yaml documents where mappings
// must be preserved, such as in certain parts of the buildkite pipeline yaml.
package ordered
